from nytid import storage
import sys
import subprocess

VALID_PERMISSIONS = "rlidwka"


class AFSfsError(Exception):
    pass


class AFSACLError(Exception):
    pass


class AFSPermissionError(Exception):
    pass


class AFSGroupError(Exception):
    pass


def set_acl(path, user_or_group, permissions):
    """
    Sets the access control list for directory `path` for user or group
    `user_or_group` to permissions in `permissions`.

    `permissions` is a list of one-letter permissions. (A string of letters and a
    list of one-letter strings are equivalent.)
    """
    valid_permissions = ""
    invalid_permissions = ""

    for permission in permissions:
        if permission in VALID_PERMISSIONS:
            valid_permissions += permission
        else:
            invalid_permissions += permission

    if invalid_permissions:
        raise AFSPermissionError(f"{invalid_permissions} are invalid permissions.")
    cmd = ["fs", "sa", path, user_or_group, permissions]
    try:
        fssa = subprocess.run(cmd, check=True, capture_output=True)
    except subprocess.CalledProcessError as err:
        errmsg = str(err.stderr, encoding=sys.getdefaultencoding())
        raise AFSACLError(
            f"Error while setting AFS ACL: " f"`{err.cmd}` resulted in '{errmsg}'"
        )


def revoke_all(path, user_or_group):
    """
    Revokes all access to `path` for `user_or_group`.
    """
    set_acl(path, user_or_group, "")


def grant_dropbox(path, user_or_group):
    """
    Lets `user_or_group` use `path` as a dropbox, i.e. can insert files, but not
    list, read or modify them.
    """
    set_acl(path, user_or_group, "i")


def grant_lookup(path, user_or_group):
    """
    Lets `user_or_group` list/lookup files in `path`.
    """
    set_acl(path, user_or_group, "l")


def grant_reader(path, user_or_group):
    """
    Lets `user_or_group` list, read and lock files in `path`.
    """
    set_acl(path, user_or_group, "rlk")


def grant_writer(path, user_or_group):
    """
    Lets `user_or_group` list, read, lock, write, delete files in `path`.
    """
    set_acl(path, user_or_group, "rlidwk")


def grant_admin(path, user_or_group):
    """
    Lets `user_or_group` do anything to `path` and files in it.
    """
    set_acl(path, user_or_group, VALID_PERMISSIONS)


def get_acl(path):
    """
    Returns a dictionary containing the access control list:
    users or groups as key, permissions as value.
    """
    acl = {}
    cmd = ["fs", "la", path]
    try:
        fsla = subprocess.run(cmd, check=True, capture_output=True)
    except subprocess.CalledProcessError as err:
        errmsg = str(err.stderr, encoding=sys.getdefaultencoding())
        raise AFSfsError(f"`fs la {path}` returned with error: {errmsg}")
    lines = str(fsla.stdout, encoding=sys.getdefaultencoding()).split("\n")
    pos_permissions = {}
    neg_permissions = {}

    while lines:
        line = lines.pop(0)
        if line == "Normal rights:":
            pos_permissions.update(pop_permissions(lines))
        elif line == "Negative rights:":
            neg_permissions.update(pop_permissions(lines))

    for key, value in pos_permissions.items():
        try:
            acl[key] = (value, neg_permissions[key])
        except KeyError:
            acl[key] = (value, None)

    for key, value in neg_permissions.items():
        if key in acl:
            continue
        else:
            acl[key] = (None, value)
    return acl


def pop_permissions(lines):
    """
    Pops all indented lines from front in `lines`. Returns dictionary containing
    username (or group) as keys, permissions as values.
    """
    acl = {}

    while lines:
        if not lines[0].startswith("  "):
            return acl

        user_or_group, permissions = lines.pop(0).split()

        acl[user_or_group] = permissions

    return acl


def create_group(group_name):
    """
    Creates a group in AFS named `group_name`.
    """
    cmd = ["pts", "creategroup", group_name]
    try:
        pts = subprocess.run(cmd, check=True, capture_output=True)
    except subprocess.CalledProcessError as err:
        pts_lines = str(err.stderr, encoding=sys.getdefaultencoding()).split("\n")
        raise AFSGroupError(f"`{' '.join(cmd)}` returned an error: {pts_lines[0]}")
    else:
        pts_lines = str(pts.stdout, encoding=sys.getdefaultencoding()).split("\n")


def add_user_to_group(username, group_name):
    """
    Adds user `username` to the AFS group `group_name`.
    """
    cmd = ["pts", "adduser", username, group_name]
    try:
        pts = subprocess.run(cmd, check=True, capture_output=True)
    except subprocess.CalledProcessError as err:
        pts_lines = str(err.stderr, encoding=sys.getdefaultencoding()).split("\n")
        raise AFSGroupError(f"`{' '.join(cmd)}` returned an error: {pts_lines[0]}")
    else:
        pts_lines = str(pts.stdout, encoding=sys.getdefaultencoding()).split("\n")


def list_users_in_group(group_name):
    """
    Lists the members of the group `group_name`.
    """
    cmd = ["pts", "membership", group_name]
    try:
        pts = subprocess.run(cmd, check=True, capture_output=True)
    except subprocess.CalledProcessError as err:
        pts_lines = str(err.stderr, encoding=sys.getdefaultencoding()).split("\n")
        raise AFSGroupError(f"`{' '.join(cmd)}` returned an error: {pts_lines[0]}")
    else:
        pts_lines = str(pts.stdout, encoding=sys.getdefaultencoding()).split("\n")

    first_line = pts_lines.pop(0)
    if first_line.startswith("pts: "):
        raise AFSGroupError(f"`{' '.join(cmd)}` returned an error: {first_line}")

    return [user.strip() for user in pts_lines]


def remove_user_from_group(username, group_name):
    """
    Removes user `username` from the AFS group `group_name`.
    """
    cmd = ["pts", "removeuser", username, group_name]
    try:
        pts = subprocess.run(cmd, check=True, capture_output=True)
    except subprocess.CalledProcessError as err:
        pts_lines = str(err.stderr, encoding=sys.getdefaultencoding()).split("\n")
        raise AFSGroupError(f"`{' '.join(cmd)}` returned an error: {pts_lines[0]}")
    else:
        pts_lines = str(pts.stdout, encoding=sys.getdefaultencoding()).split("\n")


def delete_group(group_name):
    """
    Deletes the AFS group `group_name`.
    """
    cmd = ["pts", "delete", group_name]
    try:
        pts = subprocess.run(cmd, check=True, capture_output=True)
    except subprocess.CalledProcessError as err:
        pts_lines = str(err.stderr, encoding=sys.getdefaultencoding()).split("\n")
        raise AFSGroupError(f"`{' '.join(cmd)}` returned an error: {pts_lines[0]}")
    else:
        pts_lines = str(pts.stdout, encoding=sys.getdefaultencoding()).split("\n")
    pts_lines = str(pts.stderr, encoding=sys.getdefaultencoding()).split("\n")
    if pts_lines[0].startswith("pts: "):
        raise AFSGroupError(f"`{' '.join(cmd)}` returned an error: {pts_lines[0]}")


class StorageRoot(storage.StorageRoot):
    """
    Manages a storage root in the AFS system.
    """

    def revoke_access(self, user_or_group):
        """
        Revokes access to storage root for a user or group. Returns nothing.
        """
        revoke_all(self.path, user_or_group)

    def grant_access(self, user_or_group, permissions):
        """
        Sets `permissions` as access rights for `user_or_group`.

        `permissions` is a substring of the AFS permissions: "rlidwka".

        l -- Lookup: Note that a user needs Lookup to a parent directory
                     as well as access to a subdirectory to access files in
                     the subdirectory.
        r -- Read:   Allows user to read or copy any file in the directory.
        w -- Write:  Allows user to modify any existing file in the directory.
        k -- Lock:   Allows user to flock a file.
        i -- Insert: Allows user to add new files and create new subdirectories.
        d -- Delete: Allows user to remove files and empty subdirectories.
        a -- Admin:  Allows user to change the ACL for a directory.
        """
        set_acl(self.path, user_or_group, permissions)



import arrow
import csv
import datetime
from enum import Enum
import ics.icalendar
import json
import logging
import pathlib
import os
import sys
import typer
from typing_extensions import Annotated

from nytid.signup import hr
from nytid.signup import sheets
import operator

from nytid.cli import courses as coursescli
from nytid.cli.signupsheets import SIGNUPSHEET_URL_PATH
from nytid import courses as courseutils
from nytid import schedules as schedutils
from nytid.signup import hr
from nytid.signup import sheets

import appdirs

try:
    import canvasapi
    import canvaslms.cli

    dirs_canvas = appdirs.AppDirs("canvaslms", "dbosk@kth.se")

    canvaslms_config = canvaslms.cli.read_configuration(
        f"{dirs_canvas.user_config_dir}/config.json"
    )

    CANVAS_SERVER, CANVAS_TOKEN = canvaslms.cli.login.load_credentials(canvaslms_config)

    if CANVAS_SERVER and CANVAS_TOKEN:
        canvas_session = canvasapi.Canvas(
            os.environ["CANVAS_SERVER"], os.environ["CANVAS_TOKEN"]
        )
    else:
        canvas_session = None
        logging.warning("Can't load Canvas credentials, run `canvaslms login`")
except ImportError as err:
    logging.warning(f"Can't import Canvas: {err}")
    canvas_session = None
except Exception as err:
    logging.warning(f"Can't load Canvas credentials: {err}")
    canvas_session = None
try:
    import ladok3
    import ladok3.cli

    dirs_ladok = appdirs.AppDirs("ladok", "dbosk@kth.se")

    LADOK_INST, LADOK_VARS = ladok3.cli.load_credentials(
        "f{dirs_ladok.user_config_dir}/config.json"
    )

    if LADOK_INST and LADOK_VARS:
        ladok_session = ladok3.LadokSession(LADOK_INST, LADOK_VARS)
    else:
        ladok_session = None
        logging.warning("Can't load LADOK credentials, run `ladok login`")
except ImportError as err:
    logging.warning(f"Can't import ladok3, not using LADOK data: {err}")
    ladok_session = None
except Exception as err:
    logging.warning(f"Can't load LADOK credentials: {err}")
    ladok_session = None
import cachetools
from nytid.signup import sheets
from nytid.signup import hr
import re
import typerconf
from nytid.signup.hr import timesheet

AMANUENSIS_CONTRACT_PATH = "hr.amanuensis.contract_dir"
TIMESHEETS_DIR_PATH = "hr.timesheets.timesheets_dir"

cli = typer.Typer(name="hr", help="Manage sign-up sheets for teaching")


@cachetools.cached(cache={})
def get_canvas_courses(course_regex):
    """
    Returns a list of Canvas course objects matching the given course_regex.
    """
    courses = list(canvaslms.cli.courses.filter_courses(canvas_session, course_regex))
    return courses


@cachetools.cached(cache={})
def get_canvas_users(username_regex, course_regex):
    """
    Returns a list of Canvas user objects matching the given username_regex.
    Searches for username_regex in the courses matching course_regex.
    """
    courses = get_canvas_courses(course_regex)
    users = list(canvaslms.cli.users.filter_users(courses, username_regex))
    return users


def get_canvas_user(username, course_regex):
    """
    Takes a username and returns a Canvas user object.
    Searches for username in the courses matching course_regex.
    """
    users = get_canvas_users(".*", course_regex)
    username = username.strip()
    users_without_login_id = []

    for user in users:
        try:
            if user.login_id.split("@")[0] == username or user.login_id == username:
                return user
        except AttributeError:
            users_without_login_id.append(user)
    users_without_login_id_str = "\n  ".join(map(str, users_without_login_id))
    raise ValueError(
        f"Can't find {username} in Canvas, but the following users "
        f"don't have a login ID: {users_without_login_id_str}"
    )


def get_ladok_user(canvas_user):
    """
    Takes a Canvas user object and returns a LADOK student object.
    """
    try:
        return ladok_session.get_student(canvas_user.integration_id)
    except KeyError as err:
        raise KeyError(f"can't look up {canvas_user} in LADOK: {err}")


def to_hours(td):
    return td.total_seconds() / 60 / 60


def push_forward(start, end, push_start):
    """
    Takes a start and end date and pushes them forward so that start becomes
    push_start.
    """
    if push_start > start:
        end += push_start - start
        start = push_start

    return start, end


def get_valid_contracts(user_regex):
    """
    Returns a list of valid contracts matching `user_regex`.
    """
    contracts = []

    try:
        contracts_path = pathlib.Path(typerconf.get(AMANUENSIS_CONTRACT_PATH))
    except KeyError as err:
        logging.warning(
            f"Can't find {AMANUENSIS_CONTRACT_PATH} in config, "
            f"looking for contract data in `./`. Set by running "
            f"`nytid config {AMANUENSIS_CONTRACT_PATH} -s <path>`."
        )
        contracts_path = pathlib.Path("./")
    users = set()
    user_pattern = re.compile(user_regex)
    for contract in contracts_path.glob("*.json"):
        username = contract.name.split(".")[0]
        if user_pattern.match(username):
            users.add(username)
    for user in users:
        user_contracts = []

        contract_files = sorted(
            contracts_path.glob(f"{user}.*.json"), key=lambda x: x.name
        )
        prev_contract = None
        for contract_file in contract_files:
            try:
                with open(contract_file) as infile:
                    contract = json.load(infile)
            except Exception as err:
                logging.warning(f"Can't read {contract_file}, skipping: {err}")
                continue

            if prev_contract:
                if not contract_overlap(prev_contract, contract):
                    user_contracts.append(prev_contract)

            prev_contract = contract

        user_contracts.append(contract)

        contracts += user_contracts

    return contracts


def contract_overlap(contract1, contract2):
    """
    Returns True if the contracts overlap, False otherwise.

    We define overlap as having at least one event in common.
    """
    for event1 in contract1["events"]:
        for event2 in contract2["events"]:
            if event_overlap(event1, event2):
                return True
    return False


def event_overlap(event1, event2):
    """
    Returns True if the events overlap, False otherwise.

    We define overlap as overlapping in time.
    """
    start1 = datetime.datetime.fromisoformat(event1[1])
    end1 = datetime.datetime.fromisoformat(event1[2])
    start2 = datetime.datetime.fromisoformat(event2[1])
    end2 = datetime.datetime.fromisoformat(event2[2])

    return start1 < end2 and start2 < end1


def remove_events(events, events_to_remove):
    """
    Removes the events in `events_to_remove` from `events`.
    """
    return list(filter(lambda x: x not in events_to_remove, events))


def salary_difference(added_events, removed_events):
    """
    Returns the difference in salary between the added and removed events.
    """
    return sum(map(lambda x: x["belopp"], added_events)) - sum(
        map(lambda x: x["belopp"], removed_events)
    )


def summarize_user(user, course_events, salary=150):
    """
    Returns events where TA worked.
    - `user` is the TA's username.
    - `course_events` is a list of events.
    - Optional `salary` is the hourly salary.
    """
    start_idx = sheets.SIGNUP_SHEET_HEADER.index("Start")
    end_idx = sheets.SIGNUP_SHEET_HEADER.index("End")
    type_idx = sheets.SIGNUP_SHEET_HEADER.index("Event")
    try:
        hours = to_hours(hr.hours_per_TA(course_events)[user])
    except KeyError:
        hours = 0
    events = sheets.filter_events_by_TA(
        user, sorted(course_events, key=operator.itemgetter(start_idx))
    )
    events = filter(lambda x: user in sheets.get_booked_TAs_from_csv(x)[0], events)
    events = list(
        map(lambda x: x[0 : len(sheets.SIGNUP_SHEET_HEADER)] + [user], events)
    )

    xl_events = []

    for event in events:
        end = arrow.get(event[end_idx])
        start = arrow.get(event[start_idx])
        event_type = event[type_idx]
        time = end - start
        time_with_prep = to_hours(
            hr.round_time(hr.add_prep_time(time, event_type, date=start.date()))
        )
        course = ""
        codes = re.findall(r"[A-Z]{2,4}\d{3,4}[A-Z]?", event_type)
        if codes:
            course = ", ".join(codes)

        xl_events.append(
            {
                "datum": str(start.date()),
                "tid": str(start.time()),
                "kurskod": course,
                "typ": event_type,
                "timmar": to_hours(time),
                "koeff": hr.prep_factor(
                    event_type, date=start.date(), amanuensis=False
                ),
                "omr_tid": time_with_prep,
                "belopp": time_with_prep * salary,
            }
        )

    return xl_events


detailed_opt = typer.Option(help="Output detailed user data.")
course_summary_opt = typer.Option(help="Print a summary of the course.")
amanuensis_summary_opt = typer.Option(help="Print a summary of the " "amanuensis.")
hourly_summary_opt = typer.Option(help="Print a summary of the hourly TAs.")
draft_opt = typer.Option(help="Create a draft, " "don't store the generated contract.")
user_regex_opt = typer.Option(
    "--user", help="Regex to match TAs' usernames that " "should be included."
)
start_date_opt = typer.Option(
    help="The start date (inclusive, <=), "
    "when unset includes "
    "everything in the sign-up sheet. "
    "Set this to decide what to include from "
    "the sign-up sheet.",
    formats=["%Y-%m-%d"],
)
end_date_opt = typer.Option(
    help="The end date (not inclusive, <), "
    "when unset includes "
    "everything in the sign-up sheet. "
    "Set this to decide what to include from "
    "the sign-up sheet.",
    formats=["%Y-%m-%d"],
)
delimiter_opt = typer.Option(help="Delimiter to use in CSV output.")
event_summary_opt = typer.Option(help="Print a summary of the hours per event " "type.")
push_start_opt = typer.Option(
    help="Push the dates of the contract so that it "
    "starts at this date. "
    "This keeps the same percentage.",
    formats=["%Y-%m-%d"],
)
set_start_opt = typer.Option(
    help="Force the start date of the contract to "
    "this date. Might modify percentage.",
    formats=["%Y-%m-%d"],
)
set_end_opt = typer.Option(
    help="Force the end date of the contract to " "this date. Might modify percentage.",
    formats=["%Y-%m-%d"],
)
user_regex_arg = typer.Argument(
    help="Regex to match TAs' usernames that " "should be included."
)
events_opt = typer.Option(help="Print the events used to compute the " "contract.")
updates_opt = typer.Option(
    help="Also show a draft contract based on the " "current bookings."
)
diff_opt = typer.Option(
    help="diff: Only generate time sheets when there's a "
    "difference in salary. "
    "no-diff: Always generate time sheets when there "
    "are changes, even if there is no change in "
    "salary to be paid. "
    "(E.g. it's a change of dates.) "
    "No time sheets will be generated if there "
    "are no new events."
)
draft_opt = typer.Option(
    help="Print a draft of the time sheet to stdout. " "Don't generate any timesheet."
)
store_opt = typer.Option(help="Store the time sheet data as reported.")
amanuensis_opt = typer.Option(
    help="Include only amanuenses, i.e. generate "
    "mertid time sheets. no-amanuensis "
    "include only hourly TAs."
)
sign_opt = typer.Option(help="Include the course responsible's signature.")
course_responsible_opt = typer.Option(
    help="The course responsible's name; "
    "If no default is set, we try to extract the value from the courses' config. "
    "You can set the default by running `nytid config ...` or "
    "`nytid courses config <course> me.name -s <name>`."
)
try:
    default_course_responsible = typerconf.get("me.name")
except KeyError:
    default_course_responsible = ""
course_responsible_email_opt = typer.Option(
    help="The course responsible's "
    "email; "
    "If no default is set, we try to extract the value from the courses' config. "
    "You can set the default by running `nytid config ...` or "
    "`nytid courses config <course> me.email -s <email>`."
)
try:
    default_email = typerconf.get("me.email")
except KeyError:
    default_email = ""
course_responsible_signature_opt = typer.Option(
    help="Path to a picture of "
    "the course "
    "responsible's "
    "signature; "
    "If no default is set, we try to extract the value from the courses' config. "
    "You can set the default by running `nytid config ...` or "
    "`nytid courses config <course> me.signature -s <path>`."
)
try:
    default_signature_file = pathlib.Path(typerconf.get("me.signature"))
except KeyError:
    default_signature_file = None
manager_opt = typer.Option(
    help="The manager's name; "
    "If no default is set, we try to extract the value from the courses' config. "
    "You can set the default by running `nytid config ...` or "
    "`nytid courses config <course> hr.manager -s <name>`."
)
try:
    default_manager = typerconf.get("hr.manager")
except KeyError:
    default_manager = ""
org_opt = typer.Option(
    help="The organization code for accounting; "
    "If no default is set, we try to extract the value from the courses' config. "
    "You can set the default by running `nytid config ...` or "
    "`nytid courses config <course> hr.organization -s <code>`."
)
try:
    default_organization = typerconf.get("hr.organization")
except KeyError:
    default_organization = ""
project_opt = typer.Option(
    help="The project number for accounting; "
    "If no default is set, we try to extract the value from the courses' config. "
    "You can set the default by running `nytid config ...` or "
    "`nytid courses config <course> hr.project -s <code>`."
)
try:
    default_project = typerconf.get("hr.project")
except KeyError:
    default_project = ""
reverse_opt = typer.Option(help="Reverse the sort order.")
all_opt = typer.Option(
    "--all/--newest", help="Show all time sheets or just the newest."
)
date_regex_opt = typer.Option(
    help="Filter time sheets by date. This is a "
    "regex matching the date of the time "
    "sheet."
)


@cli.command()
def users(
    course_regex: Annotated[str, coursescli.course_arg_regex],
    register: Annotated[str, coursescli.register_opt_regex] = coursescli.MINE,
    detailed: Annotated[bool, detailed_opt] = False,
):
    """
    Prints the list of all usernames booked on the course.
    """
    registers = coursescli.registers_regex(register)
    courses = {}
    for course_reg in coursescli.courses_regex(course_regex, registers):
        try:
            courses[course_reg] = courseutils.get_course_config(*course_reg)
        except KeyError as err:
            logging.warning(err)
        except PermissionError as err:
            course, register = course_reg
            logging.warning(f"You don't have access to {course} in {register}: {err}")
    if not courses:
        sys.exit(1)
    booked = []
    for (course, register), config in courses.items():
        try:
            url = config.get(SIGNUPSHEET_URL_PATH)
        except KeyError as err:
            logging.warning(
                f"Can't find sign-up sheet URL for {course} in {register}: " f"{err}"
            )
            continue
        if "docs.google.com" in url:
            url = sheets.google_sheet_to_csv_url(url)
        booked += sheets.read_signup_sheet_from_url(url)

    for user in hr.hours_per_TA(booked):
        if detailed:
            user_obj = user
            try:
                user_obj = get_canvas_user(user, course_regex)
            except Exception as err:
                logging.warning(f"Can't look up {user} in Canvas: {err}")
            else:
                try:
                    email = user_obj.email
                    user_obj = get_ladok_user(user_obj)
                    user_obj.email = email
                except Exception as err:
                    logging.warning(
                        f"Can't look up {user} ({user_obj}) in LADOK: {err}"
                    )
                    pass
        else:
            user_obj = user
        try:
            if detailed and "@" not in str(user_obj):
                print(f"{user_obj} <{user_obj.email}>")
            else:
                print(user_obj)
        except AttributeError:
            print(user_obj)


@cli.command()
def time(
    course_regex: Annotated[str, coursescli.course_arg_regex],
    register: Annotated[str, coursescli.register_opt_regex] = coursescli.MINE,
    detailed: Annotated[bool, detailed_opt] = False,
    course_summary: Annotated[bool, course_summary_opt] = True,
    amanuensis_summary: Annotated[bool, amanuensis_summary_opt] = True,
    hourly_summary: Annotated[bool, hourly_summary_opt] = True,
):
    """
    Summarizes the time spent on teaching the course(s).
    """
    registers = coursescli.registers_regex(register)
    courses = {}
    for course_reg in coursescli.courses_regex(course_regex, registers):
        try:
            courses[course_reg] = courseutils.get_course_config(*course_reg)
        except KeyError as err:
            logging.warning(err)
        except PermissionError as err:
            course, register = course_reg
            logging.warning(f"You don't have access to {course} in {register}: {err}")
    if not courses:
        sys.exit(1)
    booked = []
    for (course, register), config in courses.items():
        try:
            url = config.get(SIGNUPSHEET_URL_PATH)
        except KeyError as err:
            logging.warning(
                f"Can't find sign-up sheet URL for {course} in {register}: " f"{err}"
            )
            continue
        if "docs.google.com" in url:
            url = sheets.google_sheet_to_csv_url(url)
        booked += sheets.read_signup_sheet_from_url(url)
    csvout = csv.writer(sys.stdout, delimiter="\t")
    if course_summary:
        h_per_student = hr.hours_per_student(booked)

        for event, hours in h_per_student.items():
            csvout.writerow([event, to_hours(hours), "h/student"])

        csvout.writerow(
            [
                "Booked (h)",
                to_hours(hr.total_hours(booked)),
                to_hours(hr.max_hours(booked)),
            ]
        )
    if amanuensis_summary:
        if course_summary:
            csvout.writerow([])
        if hourly_summary:
            csvout.writerow(["# Amanuensis"])

        amanuensis = hr.compute_amanuensis_data(booked)

        for user, data in amanuensis.items():
            if not user:
                continue
            if detailed:
                user_obj = user
                try:
                    user_obj = get_canvas_user(user, course_regex)
                except Exception as err:
                    logging.warning(f"Can't look up {user} in Canvas: {err}")
                else:
                    try:
                        email = user_obj.email
                        user_obj = get_ladok_user(user_obj)
                        user_obj.email = email
                    except Exception as err:
                        logging.warning(
                            f"Can't look up {user} ({user_obj}) in LADOK: {err}"
                        )
                        pass
            else:
                user_obj = user
            csvout.writerow(
                [
                    user_obj,
                    f"{data[2]:.2f} h",
                    f"{100*hr.compute_percentage(*data):.1f}%",
                    f"{data[0].format('YYYY-MM-DD')}",
                    f"{data[1].format('YYYY-MM-DD')}",
                ]
            )
    if hourly_summary:
        if amanuensis_summary:
            csvout.writerow([])
            csvout.writerow(["# Hourly"])
        elif course_summary:
            csvout.writerow([])

        for user, hours in hr.hours_per_TA(booked).items():
            if not user or user in amanuensis:
                continue
            if detailed:
                user_obj = user
                try:
                    user_obj = get_canvas_user(user, course_regex)
                except Exception as err:
                    logging.warning(f"Can't look up {user} in Canvas: {err}")
                else:
                    try:
                        email = user_obj.email
                        user_obj = get_ladok_user(user_obj)
                        user_obj.email = email
                    except Exception as err:
                        logging.warning(
                            f"Can't look up {user} ({user_obj}) in LADOK: {err}"
                        )
                        pass
            else:
                user_obj = user

            try:
                user_text = (
                    f"{user_obj} <{user_obj.email}>"
                    if "@" not in str(user_obj)
                    else str(user_obj)
                )
            except AttributeError:
                user_text = str(user_obj)

            csvout.writerow([user_text, to_hours(hours), "h"])


amanuensis = typer.Typer(name="amanuensis", help="Manage amanuensis employment")
cli.add_typer(amanuensis)


@amanuensis.command(name="create")
def cli_amanuens_create(
    user_regex: Annotated[str, user_regex_opt] = ".*",
    start: Annotated[datetime.datetime, start_date_opt] = None,
    end: Annotated[datetime.datetime, end_date_opt] = None,
    push_start: Annotated[datetime.datetime, push_start_opt] = None,
    set_start: Annotated[datetime.datetime, set_start_opt] = None,
    set_end: Annotated[datetime.datetime, set_end_opt] = None,
    course_regex: Annotated[str, coursescli.course_arg_regex] = ".*",
    register: Annotated[str, coursescli.register_opt_regex] = coursescli.MINE,
    detailed: Annotated[bool, detailed_opt] = True,
    event_summary: Annotated[bool, event_summary_opt] = False,
    draft: Annotated[bool, draft_opt] = False,
    delimiter: Annotated[str, delimiter_opt] = "\t",
):
    """
    Computes amanuensis data for a TA.
    """
    if start:
        start = start.astimezone()
    if push_start:
        push_start = push_start.astimezone()
    if set_start:
        set_start = set_start.astimezone()
    if end:
        end = end.astimezone()
    if set_end:
        set_end = set_end.astimezone()
    registers = coursescli.registers_regex(register)
    courses = {}
    for course_reg in coursescli.courses_regex(course_regex, registers):
        try:
            courses[course_reg] = courseutils.get_course_config(*course_reg)
        except KeyError as err:
            logging.warning(err)
        except PermissionError as err:
            course, register = course_reg
            logging.warning(f"You don't have access to {course} in {register}: {err}")
    if not courses:
        sys.exit(1)
    booked = []
    for (course, register), config in courses.items():
        try:
            url = config.get(SIGNUPSHEET_URL_PATH)
        except KeyError as err:
            logging.warning(
                f"Can't find sign-up sheet URL for {course} in {register}: " f"{err}"
            )
            continue
        if "docs.google.com" in url:
            url = sheets.google_sheet_to_csv_url(url)
        booked += sheets.read_signup_sheet_from_url(url)
    booked = sheets.filter_events_by_date(booked, start, end)

    amanuensis = hr.compute_amanuensis_data(booked, begin_date=start, end_date=end)

    user_pattern = re.compile(user_regex)
    first_print = True
    csvout = csv.writer(sys.stdout, delimiter=delimiter)
    path = pathlib.Path("./")
    try:
        path = pathlib.Path(typerconf.get(AMANUENSIS_CONTRACT_PATH))
    except KeyError as err:
        logging.warning(
            f"Can't find {AMANUENSIS_CONTRACT_PATH} in config, "
            f"storing contract data in `{path}`. Set by running "
            f"`nytid config {AMANUENSIS_CONTRACT_PATH} -s <path>`."
        )
    for user in amanuensis:
        if not user_pattern.match(user):
            continue
        if first_print:
            first_print = False
        else:
            print("\n")

        data = amanuensis[user]

        start = data[0]
        end = data[1]
        hours = data[2]
        if push_start:
            push_start = arrow.Arrow(push_start.year, push_start.month, push_start.day)
            start, end = push_forward(start, end, push_start)
        data = list(data)
        if set_start:
            start = data[0] = set_start
        if set_end:
            end = data[1] = set_end

        if detailed:
            user_obj = user
            try:
                user_obj = get_canvas_user(user, course_regex)
            except Exception as err:
                logging.warning(f"Can't look up {user} in Canvas: {err}")
            else:
                try:
                    email = user_obj.email
                    user_obj = get_ladok_user(user_obj)
                    user_obj.email = email
                except Exception as err:
                    logging.warning(
                        f"Can't look up {user} ({user_obj}) in LADOK: {err}"
                    )
                    pass
        else:
            user_obj = user

        try:
            user_text = (
                f"{user_obj} <{user_obj.email}>"
                if "@" not in str(user_obj)
                else str(user_obj)
            )
        except AttributeError:
            user_text = str(user_obj)

        row = [
            user_text,
            start.date(),
            end.date(),
            f"{round(100*hr.compute_percentage(*data))}%",
        ]

        if event_summary:
            row.append(f"{hours:.2f} h")

        csvout.writerow(row)

        user_events = sheets.filter_events_by_TA(user, booked)
        user_events = filter(
            lambda x: user in sheets.get_booked_TAs_from_csv(x)[0], booked
        )
        user_events = list(
            map(lambda x: x[0 : len(sheets.SIGNUP_SHEET_HEADER)] + [user], user_events)
        )
        if event_summary:
            for event, hours in hr.hours_per_event(user_events).items():
                csvout.writerow([event, to_hours(hours), "h"])
        if not draft:
            filename = f"{user}.{datetime.datetime.now().isoformat()}.json"

            path.mkdir(parents=True, exist_ok=True)

            with open(path / filename, "w") as outfile:
                json.dump(
                    {
                        "user": user,
                        "start": start.isoformat() if start else None,
                        "set_start": set_start.isoformat() if set_start else None,
                        "push_start": push_start.isoformat() if push_start else None,
                        "end": end.isoformat() if end else None,
                        "set_end": set_end.isoformat() if set_end else None,
                        "course_regex": course_regex,
                        "events": user_events,
                    },
                    outfile,
                    indent=2,
                )


@amanuensis.command(name="show")
def cli_amanuens_show(
    user_regex: Annotated[str, user_regex_arg] = ".*",
    detailed: Annotated[bool, detailed_opt] = False,
    event_summary: Annotated[bool, event_summary_opt] = True,
    delimiter: Annotated[str, delimiter_opt] = "\t",
    updates: Annotated[bool, updates_opt] = False,
    events: Annotated[bool, events_opt] = False,
):
    """
    Shows stored amanuesis contracts for TAs.
    """
    try:
        contracts_path = pathlib.Path(typerconf.get(AMANUENSIS_CONTRACT_PATH))
    except KeyError as err:
        logging.warning(
            f"Can't find {AMANUENSIS_CONTRACT_PATH} in config, "
            f"looking for contract data in `./`. Set by running "
            f"`nytid config {AMANUENSIS_CONTRACT_PATH} -s <path>`."
        )
        contracts_path = pathlib.Path("./")
    contracts = get_valid_contracts(user_regex)

    user_pattern = re.compile(user_regex)
    first_print = True
    csvout = csv.writer(sys.stdout, delimiter=delimiter)
    path = pathlib.Path("./")
    try:
        path = pathlib.Path(typerconf.get(AMANUENSIS_CONTRACT_PATH))
    except KeyError as err:
        logging.warning(
            f"Can't find {AMANUENSIS_CONTRACT_PATH} in config, "
            f"storing contract data in `{path}`. Set by running "
            f"`nytid config {AMANUENSIS_CONTRACT_PATH} -s <path>`."
        )
    for contract in contracts:
        user = contract["user"]
        booked = contract["events"]

        start = (
            datetime.datetime.fromisoformat(contract["start"])
            if "start" in contract and contract["start"]
            else None
        )
        push_start = (
            datetime.datetime.fromisoformat(contract["push_start"])
            if "push_start" in contract and contract["push_start"]
            else None
        )
        set_start = (
            datetime.datetime.fromisoformat(contract["set_start"])
            if "set_start" in contract and contract["set_start"]
            else None
        )
        end = (
            datetime.datetime.fromisoformat(contract["end"])
            if "end" in contract and contract["end"]
            else None
        )
        set_end = (
            datetime.datetime.fromisoformat(contract["set_end"])
            if "set_end" in contract and contract["set_end"]
            else None
        )

        amanuensis = hr.compute_amanuensis_data(booked, begin_date=start, end_date=end)

        try:
            course_regex = contract["course_regex"]
        except KeyError:
            code_pattern = re.compile(r"[A-Z]{2,4}\d{3,4}[A-Z]?")
            for event in booked:
                match = code_pattern.search(event[0])
                if match:
                    course_regex = match.group(0)
                    break
            else:
                course_regex = ".*"

        if first_print:
            first_print = False
        else:
            print("\n")

        data = amanuensis[user]

        start = data[0]
        end = data[1]
        hours = data[2]
        if push_start:
            push_start = arrow.Arrow(push_start.year, push_start.month, push_start.day)
            start, end = push_forward(start, end, push_start)
        data = list(data)
        if set_start:
            start = data[0] = set_start
        if set_end:
            end = data[1] = set_end

        if detailed:
            user_obj = user
            try:
                user_obj = get_canvas_user(user, course_regex)
            except Exception as err:
                logging.warning(f"Can't look up {user} in Canvas: {err}")
            else:
                try:
                    email = user_obj.email
                    user_obj = get_ladok_user(user_obj)
                    user_obj.email = email
                except Exception as err:
                    logging.warning(
                        f"Can't look up {user} ({user_obj}) in LADOK: {err}"
                    )
                    pass
        else:
            user_obj = user

        try:
            user_text = (
                f"{user_obj} <{user_obj.email}>"
                if "@" not in str(user_obj)
                else str(user_obj)
            )
        except AttributeError:
            user_text = str(user_obj)

        row = [
            user_text,
            start.date(),
            end.date(),
            f"{round(100*hr.compute_percentage(*data))}%",
        ]

        if event_summary:
            row.append(f"{hours:.2f} h")

        csvout.writerow(row)

        user_events = sheets.filter_events_by_TA(user, booked)
        user_events = filter(
            lambda x: user in sheets.get_booked_TAs_from_csv(x)[0], booked
        )
        user_events = list(
            map(lambda x: x[0 : len(sheets.SIGNUP_SHEET_HEADER)] + [user], user_events)
        )
        if event_summary:
            for event, hours in hr.hours_per_event(user_events).items():
                csvout.writerow([event, to_hours(hours), "h"])
        if events:
            print()
            csvout = csv.writer(sys.stdout, delimiter="\t")
            for event in contract["events"]:
                csvout.writerow(event)
        if updates:
            print()
            try:
                start = (
                    datetime.datetime.fromisoformat(contract["start"])
                    if contract["start"]
                    else None
                )
            except KeyError:
                start = None

            try:
                end = (
                    datetime.datetime.fromisoformat(contract["end"])
                    if contract["end"]
                    else None
                )
            except KeyError:
                end = None

            try:
                push_start = (
                    datetime.datetime.fromisoformat(contract["push_start"])
                    if contract["push_start"]
                    else None
                )
            except KeyError:
                push_start = None

            try:
                set_start = (
                    datetime.datetime.fromisoformat(contract["set_start"])
                    if contract["set_start"]
                    else None
                )
            except KeyError:
                set_start = None

            try:
                set_end = (
                    datetime.datetime.fromisoformat(contract["set_end"])
                    if contract["set_end"]
                    else None
                )
            except KeyError:
                set_end = None

            cli_amanuens_create(
                user_regex=contract["user"],
                course_regex=".*",
                start=start,
                end=end,
                push_start=push_start,
                set_start=set_start,
                set_end=set_end,
                detailed=detailed,
                event_summary=event_summary,
                draft=True,
                delimiter=delimiter,
            )


timesheets = typer.Typer(name="timesheets", help="Manage timesheets for TAs")
cli.add_typer(timesheets)


@timesheets.command(name="generate")
def cli_generate_timesheets(
    user_regex: Annotated[str, user_regex_opt] = ".*",
    start: Annotated[datetime.datetime, start_date_opt] = None,
    end: Annotated[datetime.datetime, end_date_opt] = None,
    course_responsible: Annotated[
        str, course_responsible_opt
    ] = default_course_responsible,
    course_responsible_email: Annotated[
        str, course_responsible_email_opt
    ] = default_email,
    course_responsible_signature: Annotated[
        pathlib.Path, course_responsible_signature_opt
    ] = default_signature_file,
    manager: Annotated[str, manager_opt] = default_manager,
    organization: Annotated[str, org_opt] = default_organization,
    project: Annotated[str, project_opt] = default_project,
    course_regex: Annotated[str, coursescli.course_arg_regex] = ".*",
    register: Annotated[str, coursescli.register_opt_regex] = coursescli.MINE,
    store: Annotated[bool, store_opt] = True,
    amanuensis: Annotated[bool, amanuensis_opt] = False,
    sign: Annotated[bool, sign_opt] = True,
    diff: Annotated[bool, diff_opt] = True,
    draft: Annotated[bool, draft_opt] = False,
):
    """
    Generates time sheets for TAs in courses.
    """
    registers = coursescli.registers_regex(register)
    courses = {}
    for course_reg in coursescli.courses_regex(course_regex, registers):
        try:
            courses[course_reg] = courseutils.get_course_config(*course_reg)
        except KeyError as err:
            logging.warning(err)
        except PermissionError as err:
            course, register = course_reg
            logging.warning(f"You don't have access to {course} in {register}: {err}")
    if not courses:
        sys.exit(1)
    managers = []
    orgs = []
    projects = []
    for _, course_config in courses.items():
        try:
            managers.append(course_config.get("hr.manager"))
        except KeyError:
            pass
        try:
            orgs.append(course_config.get("hr.organization"))
        except KeyError:
            pass
        try:
            projects.append(course_config.get("hr.project"))
        except KeyError:
            pass

    if len(set(managers)) == 1:
        manager = managers[0]
    if len(set(orgs)) == 1:
        organization = orgs[0]
    if len(set(projects)) == 1:
        project = projects[0]
    booked = []
    for (course, register), config in courses.items():
        try:
            url = config.get(SIGNUPSHEET_URL_PATH)
        except KeyError as err:
            logging.warning(
                f"Can't find sign-up sheet URL for {course} in {register}: " f"{err}"
            )
            continue
        if "docs.google.com" in url:
            url = sheets.google_sheet_to_csv_url(url)
        booked += sheets.read_signup_sheet_from_url(url)
    booked = sheets.filter_events_by_date(booked, start, end)

    user_pattern = re.compile(user_regex)
    users = hr.hours_per_TA(booked)
    csvout = csv.writer(sys.stdout, delimiter="\t")
    first_print = True
    for user in users:
        if not user_pattern.match(user):
            continue
        contracts = get_valid_contracts(user)
        if amanuensis and not contracts:
            continue
        elif not amanuensis and contracts:
            continue
        ta_events = summarize_user(user, booked)
        all_prev_events = []

        user_contracts = get_valid_contracts(user)
        for contract in user_contracts:
            contract_events = summarize_user(
                user, sheets.filter_events_by_date(contract["events"], start, end)
            )
            all_prev_events += contract_events
        timesheets_dir = pathlib.Path(typerconf.get(TIMESHEETS_DIR_PATH))
        for timesheet_file in timesheets_dir.glob(f"{user}.*.json"):
            try:
                with open(timesheet_file) as infile:
                    timesheet_data = json.load(infile)
            except Exception as err:
                logging.warning(f"Can't read {timesheet_file}, skipping: {err}")
                continue

            all_prev_events += timesheet_data["added_events"]
            all_prev_events = remove_events(
                all_prev_events, timesheet_data["removed_events"]
            )

        added_events = remove_events(ta_events, all_prev_events)
        removed_events = remove_events(all_prev_events, ta_events)
        salary_diff = salary_difference(added_events, removed_events)

        if (diff and salary_diff != 0) or not diff:
            if draft:
                if not (added_events or removed_events):
                    continue

                if first_print:
                    first_print = False
                else:
                    csvout.writerow([])
                    csvout.writerow([])

                if added_events or removed_events:
                    csvout.writerow([user])

                if added_events and removed_events:
                    csvout.writerow(["Added events"])

                for event in added_events:
                    csvout.writerow(event.values())

                if removed_events:
                    csvout.writerow([])
                    csvout.writerow(["Removed events"])

                for event in removed_events:
                    event["timmar"] = -event["timmar"]
                    event["omr_tid"] = -event["omr_tid"]
                    event["belopp"] = -event["belopp"]
                    csvout.writerow(event.values())
            else:
                if not added_events and not removed_events:
                    logging.warning(f"No events for {user}, skipping.")
                    continue

                user_obj = user
                try:
                    user_obj = get_canvas_user(user, course_regex)
                except Exception as err:
                    logging.warning(f"Can't look up {user} in Canvas: {err}")
                else:
                    try:
                        email = user_obj.email
                        user_obj = get_ladok_user(user_obj)
                        user_obj.email = email
                    except Exception as err:
                        logging.warning(
                            f"Can't look up {user} ({user_obj}) in LADOK: {err}"
                        )
                        pass
                try:
                    personnummer = user_obj.personnummer
                    name = f"{user_obj.first_name} {user_obj.last_name}"
                except AttributeError as err:
                    logging.warning(f"can't access {user}'s LADOK data: {err}")
                    personnummer = "-"
                    name = str(user_obj)

                output_filename = f"timesheet-{user}.xlsx"

                timesheet.make_xlsx(
                    personnummer,
                    name,
                    f"{user}@kth.se" if not "@" in user else user,
                    added_events,
                    removed_events=removed_events,
                    course_leader=(course_responsible, course_responsible_email),
                    HoD=manager,
                    org=organization,
                    project=project,
                    course_leader_signature=(
                        course_responsible_signature if sign else None
                    ),
                    output=output_filename,
                )

        if store and not draft:
            timesheets_dir = pathlib.Path(typerconf.get(TIMESHEETS_DIR_PATH))
            filename = f"{user}.{datetime.datetime.now().isoformat()}.json"
            with open(timesheets_dir / filename, "w") as outfile:
                json.dump(
                    {
                        "user": user,
                        "added_events": added_events,
                        "removed_events": removed_events,
                    },
                    outfile,
                    indent=2,
                )


@timesheets.command(name="show")
def cli_show_timesheets(
    user_regex: Annotated[str, user_regex_arg] = ".*",
    reverse: Annotated[bool, reverse_opt] = False,
    all: Annotated[bool, all_opt] = False,
    delimiter: Annotated[str, delimiter_opt] = "\t",
    date_regex: Annotated[str, date_regex_opt] = None,
):
    """
    Shows stored time sheets for TAs.
    """
    timesheets_dir = pathlib.Path(typerconf.get(TIMESHEETS_DIR_PATH))

    user_pattern = re.compile(user_regex)
    if date_regex:
        date_pattern = re.compile(date_regex)
    printed_users = set()
    csvout = csv.writer(sys.stdout, delimiter=delimiter)
    first_print = True
    if not all:
        reverse = True

    timesheets = sorted(
        timesheets_dir.glob(f"*.json"), key=lambda x: x.name, reverse=reverse
    )
    for timesheet in timesheets:
        username = timesheet.name.split(".")[0]
        if not user_pattern.match(username):
            continue
        if date_regex and not date_pattern.match(timesheet.name.split(".")[1]):
            continue
        if username in printed_users and not all:
            continue
        else:
            printed_users.add(username)
        if first_print:
            first_print = False
        else:
            csvout.writerow([])
            csvout.writerow([])

        csvout.writerow([timesheet.name])

        with open(timesheet) as infile:
            timesheet_data = json.load(infile)

        added_events = timesheet_data["added_events"]
        removed_events = timesheet_data["removed_events"]

        if added_events and removed_events:
            csvout.writerow(["Added events"])

        for event in added_events:
            csvout.writerow(event.values())

        if removed_events:
            csvout.writerow([])
            csvout.writerow(["Removed events"])

        for event in removed_events:
            csvout.writerow(event.values())



import importlib
import logging
from nytid.cli import package_contents
import typer

cli = typer.Typer(
    name="utils",
    help="Various utilities",
)

modules = package_contents(__name__)
for module_name in modules:
    try:
        module = importlib.import_module(module_name)
        cli.add_typer(module.cli)
    except Exception as err:
        logging.warning(f"Trying to add {module_name} yields: {err}")
        continue

if __name__ == "__main__":
    cli()



def needed_TAs(event, group_size=12):
    """
    Takes an event and returns the number of TAs needed
    """
    num_groups = event.description.split().count("grupp")
    if num_groups == 0:
        num_groups = event.description.split().count("group")

    num_rooms = len(event.location.split(","))

    if "laboration" in event.name or "Laboration" in event.name:
        if num_groups:
            num_students = num_groups * group_size
        else:
            num_students = group_size

        num_TAs = round(num_students / 12)
    elif "Övning" in event.name or "Exercise" in event.name or "Tutorial" in event.name:
        num_TAs = num_rooms
    else:
        num_TAs = max(num_rooms, num_groups)

    return num_TAs



import csv
from datetime import datetime
import dateutil.tz
import ics.attendee
import ics.event
import nytid.schedules
import nytid.signup.utils
import os
from nytid.signup import sheets
import re
import requests

SIGNUP_SHEET_HEADER = ["Event", "Start", "End", "Rooms", "#Needed TAs"]
DATETIME_FORMAT = "YYYY-MM-DD HH:mm"
STRPTIME_FORMAT = "%Y-%m-%d %H:%M"


class EventFromCSV(ics.event.Event):
    """A class to create an ics.event.Event from an event in CSV format"""

    def __init__(self, csv_row):
        """
        Input: a row from a calendar in CSV format (e.g. the sign-up sheet).
        """
        kwargs = dict()

        kwargs["name"] = csv_row[SIGNUP_SHEET_HEADER.index("Event")]

        tz = dateutil.tz.tzlocal()

        begin = datetime.strptime(
            csv_row[SIGNUP_SHEET_HEADER.index("Start")], STRPTIME_FORMAT
        )
        begin.replace(tzinfo=tz)
        kwargs["begin"] = begin.astimezone(dateutil.tz.UTC)

        end = datetime.strptime(
            csv_row[SIGNUP_SHEET_HEADER.index("End")], STRPTIME_FORMAT
        )
        end.replace(tzinfo=tz)
        kwargs["end"] = end.astimezone(dateutil.tz.UTC)

        rooms = csv_row[SIGNUP_SHEET_HEADER.index("Rooms")]
        kwargs["location"] = rooms
        needed_TAs = csv_row[SIGNUP_SHEET_HEADER.index("#Needed TAs")]
        kwargs["description"] = f"Needed TAs: {needed_TAs}\n"

        kwargs["attendees"] = [
            ics.attendee.Attendee(f"{user}@kth.se")
            for user in csv_row[len(SIGNUP_SHEET_HEADER) :]
            if user
        ]
        booked, reserves = sheets.get_booked_TAs_from_csv(csv_row)

        kwargs["description"] += "Booked TAs:"
        if booked:
            for username in booked:
                kwargs["description"] += f" {username}"
        else:
            kwargs["description"] += " None"
        kwargs["description"] += "\n"

        if reserves:
            kwargs["description"] += "Reserve TAs:"
            for username in reserves:
                kwargs["description"] += f" {username}"
            kwargs["description"] += "\n"

        super().__init__(**kwargs)


def generate_signup_sheet(
    outfile,
    url,
    needed_TAs=nytid.signup.utils.needed_TAs,
    event_filter=nytid.schedules.event_filter,
    digital_separately=True,
):
    """
    Input:
    - outfile is a string containing the file name used for output.
    - url is the URL to the ICS-formatted calendar.
    - needed_TAs is a function computing the number of needed TAs based on the
      event. The default is the needed_TAs function in this module,
    - event_filter is a function that filters events, takes a list of events as
      argument and returns a filtered list.
    - digital_separately is a bool indicating whether to separate digital
      events from physical ones. If True, the digital events will be separated into
      their own rows in the sign-up sheet. Default is True.

    Output:
    Returns nothing. Writes output to {outfile}.csv.
    """
    try:
        out = open(outfile, "w")
    except FileNotFoundError:
        os.makedirs(os.path.dirname(outfile), exist_ok=True)
        out = open(outfile, "w")
    with out:
        csvout = csv.writer(out, delimiter="\t")
        calendar = nytid.schedules.read_calendar(url)

        max_num_TAs = 0
        rows = []

        for event in event_filter(calendar.timeline):
            num_TAs = needed_TAs(event)

            if num_TAs > max_num_TAs:
                max_num_TAs = num_TAs

            if digital_separately and "digital" in event.location.casefold():
                TAs_per_room = num_TAs // (event.location.count(",") + 1)
                event.location = re.sub(
                    r"(,\s*digital|digital,\s*|digital)",
                    "",
                    event.location,
                    flags=re.IGNORECASE,
                ).strip()
                if event.location:  # check if digital was the only room
                    rows.append(
                        [*event_to_CSV(event), max(num_TAs - TAs_per_room, 1)]
                    )  # at least one TA
                event.location = "Digital"
                rows.append([*event_to_CSV(event), max(TAs_per_room, 1)])
            else:
                rows.append([*event_to_CSV(event), num_TAs])

        csvout.writerow(
            SIGNUP_SHEET_HEADER + [f"TA username" for n in range(max_num_TAs)] + ["..."]
        )

        csvout.writerows(rows)


def event_to_CSV(event):
    """
    Input: event is an ics.event.Event object.

    Output: a list of strings containing the event's attributes in the order
    specified in SIGNUP_SHEET_HEADER.
    """
    return [
        event.name,
        event.begin.to("local").format(DATETIME_FORMAT),
        event.end.to("local").format(DATETIME_FORMAT),
        event.location,
    ]


def read_signup_sheet_from_file(filename):
    """
    Input: filename is a string containing the file name of the CSV file of the
    sign-up sheet.

    Output: All the rows of the CSV as a Python list.
    """
    with open(filename, "r") as f:
        csvfile = csv.reader(f)
        return list(filter(any, list(csvfile)[1:]))


def read_signup_sheet_from_url(url):
    """
    Input: url is a string containing the URL of the CSV file of the sign-up
    sheet.

    Output: All the rows of the CSV as a Python list.
    """
    response = requests.get(url)
    if response.status_code != 200:
        raise ValueError(response.text)

    response.encoding = response.apparent_encoding
    csvdata = response.text.splitlines()
    return list(filter(any, list(csv.reader(csvdata))[1:]))


def google_sheet_to_csv_url(share_url):
    """
    Input: The share URL of a Google Sheets sheet.

    Output: A URL that downloads (exports) the sheet in CSV format.
    """
    match = re.search("/edit.*$", share_url)
    if not match:
        raise ValueError(f"{share_url} doesn't seem like a Google Sheets URL.")

    url = share_url[: match.start()]
    return url + "/export?format=csv"


def get_TAs_from_csv(csv_row):
    """
    Input: takes a CSV data row as from a csv.reader.

    Output: returns the list of signed TAs. Ensures casefold for TA IDs.
    """
    return list(
        map(
            lambda x: x.casefold().strip(),
            filter(lambda x: x.strip(), csv_row[len(SIGNUP_SHEET_HEADER) :]),
        )
    )


def get_booked_TAs_from_csv(csv_row):
    """
    Input: takes a CSV data row as from a csv.reader.

    Output: returns the list of signed TAs, the first N, where N is the number of
    needed TAs specified in the CSV data.
    """
    TAs = get_TAs_from_csv(csv_row)
    num_needed_TAs = int(csv_row[SIGNUP_SHEET_HEADER.index("#Needed TAs")])

    return TAs[:num_needed_TAs], TAs[num_needed_TAs:]


def filter_events_by_TA(ta_id, csv_rows):
    """
    Input: ta_id is the string to (exactly) match the TAs' identifiers against;
    csv_rows is a list of CSV rows, as from csv.reader.

    Output: a list of CSV rows containing only the rows containing ta_id.
    """
    return list(filter(lambda x: ta_id.casefold() in get_TAs_from_csv(x), csv_rows))


def filter_events_by_title(event_title, csv_rows):
    """
    Input: event_title is the substring to match the event title against;
    csv_rows is a list of CSV rows, as from csv.reader.

    Output: a list of CSV rows containing only the rows with an event title
    having event_title as substring.
    """
    return list(
        filter(lambda x: event_title in x[SIGNUP_SHEET_HEADER.index("Event")], csv_rows)
    )


def filter_events_by_date(csv_rows, start_date=None, end_date=None):
    """
    Input: start_date and end_date are datetime.date objects; csv_rows is a list
    of CSV rows, as from csv.reader.

    Output: a list of CSV rows containing only the rows with an event starting
    between start_date and end_date.
    """
    filtered = csv_rows.copy()

    if start_date:
        filtered = filter(
            lambda x: start_date.date()
            <= datetime.strptime(
                x[SIGNUP_SHEET_HEADER.index("Start")], STRPTIME_FORMAT
            ).date(),
            csv_rows,
        )
    if end_date:
        filtered = filter(
            lambda x: end_date.date()
            > datetime.strptime(
                x[SIGNUP_SHEET_HEADER.index("End")], STRPTIME_FORMAT
            ).date(),
            csv_rows,
        )

    return list(filtered)



import logging
import typer
import typerconf
import typing
from typing_extensions import Annotated

import datetime
from nytid.schedules import read_calendar
import csv
import sys

cli = typer.Typer(name="rooms", help="Finding free rooms for courses")

BOOKED_ROOMS_URL = "utils.rooms.url"
INTERESTING_ROOMS = "utils.rooms.interesting_rooms"
DATE_FORMAT = "%Y-%m-%d %H:%M"


def free_rooms(ics_url: str) -> typing.List[typing.Tuple]:
    """
    Given a URL or path to a TimeEdit ICS file (`ics_url`),
    return a CSV (list of tuples) of the free rooms:

    [(start, end, unbooked_rooms), ...]

    where start and end are datetime objects and unbooked_rooms is a set of
    strings.
    """
    results = []
    all_rooms = set(typerconf.get(INTERESTING_ROOMS))

    schedule = read_calendar(ics_url)
    for event in schedule.events:
        start = event.begin.datetime
        end = event.end.datetime

        booked_rooms = event.location.split(",")
        if len(booked_rooms) == 1 and booked_rooms[0].strip() == "":
            booked_rooms = set()
        else:
            booked_rooms = set(map(lambda x: x.strip(), booked_rooms))

        # If there are no booked rooms, we can skip this event.
        if not booked_rooms:
            continue

        unbooked_rooms = all_rooms - booked_rooms
        results.append((start, end, unbooked_rooms))

    time_dict = {}
    for start, end, unbooked_rooms in results:
        if (start, end) in time_dict:
            time_dict[(start, end)] |= unbooked_rooms
        else:
            time_dict[(start, end)] = unbooked_rooms

    results = [
        (start.strftime(DATE_FORMAT), end.strftime(DATE_FORMAT), unbooked_rooms)
        for (start, end), unbooked_rooms in time_dict.items()
    ]
    results.sort()
    return results


def booked_rooms(ics_url: str) -> typing.List[typing.Tuple]:
    """
    Given a URL or path to a TimeEdit ICS file ([[ics_url]]),
    return a ics (list of tuples) of the booked rooms:

    [(start, end, booked_rooms), ...]

    where start and end are datetime objects and booked_rooms is a set of
    strings.
    """
    results = []

    schedule = read_calendar(ics_url)
    for event in schedule.events:
        start = event.begin.datetime
        end = event.end.datetime

        booked_rooms = event.location.split(",")
        if len(booked_rooms) == 1 and booked_rooms[0].strip() == "":
            booked_rooms = set()
        else:
            booked_rooms = set(map(lambda x: x.strip(), booked_rooms))

        # If there are no booked rooms, we can skip this event.
        if not booked_rooms:
            continue

        results.append((start, end, booked_rooms))

    time_dict = {}
    for start, end, unbooked_rooms in results:
        if (start, end) in time_dict:
            time_dict[(start, end)] |= unbooked_rooms
        else:
            time_dict[(start, end)] = unbooked_rooms

    results = [
        (start.strftime(DATE_FORMAT), end.strftime(DATE_FORMAT), unbooked_rooms)
        for (start, end), unbooked_rooms in time_dict.items()
    ]
    results.sort()
    return results


delimiter_opt = Annotated[
    str,
    typer.Option(
        "-d", "--delimiter", help="CSV delimiter, default tab.", show_default=False
    ),
]


@cli.command(name="set-url")
def set_url_cmd(
    url: Annotated[
        str,
        typer.Argument(
            help="URL to TimeEdit export, " "ICS format for subscribing.",
            show_default=False,
        ),
    ],
    interesting_rooms: Annotated[
        typing.List[str],
        typer.Argument(
            help="Space-separated list of "
            "rooms in the schedule, e.g.: "
            "D1 D2 D3 D37",
            show_default=False,
        ),
    ],
):
    """
    Search for all the rooms that you're interested in in TimeEdit,
    e.g. D1, D2, D3 and D37.
    Set the relevant time intervals to something future proof: you
    don't want it to be too short, then you have to update the URL
    too often. Use a relative time frame. Get the URL for subscribing,
    ICS format, make sure to select the right time frame.
    """
    try:
        typerconf.set(BOOKED_ROOMS_URL, url)
    except Exception as err:
        logging.error(f"Can't set URL: {err}")
        raise typer.Exit(1)
    try:
        typerconf.set(INTERESTING_ROOMS, interesting_rooms)
    except Exception as err:
        logging.error(f"Can't set interesting rooms: {err}")
        raise typer.Exit(1)


@cli.command(name="unbooked")
def unbooked_cmd(
    delimiter: delimiter_opt = "\t",
):
    """
    Shows date and time and which rooms are free. Note that if a time is NOT in the
    list, it means that all rooms are FREE.
    """
    try:
        rooms_url = typerconf.get(BOOKED_ROOMS_URL)
    except Exception as err:
        logging.error(f"Can't get URL from config: {err}")
        logging.info("Please set it with " "'nytid utils rooms set-url <url>'")
        raise typer.Exit(1)
    try:
        csv_out = csv.writer(sys.stdout, delimiter=delimiter)

        for start, end, rooms in free_rooms(rooms_url):
            csv_out.writerow([start, end, ", ".join(sorted(rooms))])
    except Exception as err:
        logging.error(f"Can't get free rooms: {err}")
        raise typer.Exit(1)


@cli.command(name="booked")
def booked_cmd(
    delimiter: delimiter_opt = "\t",
):
    """
    Shows date and time and which rooms are booked.
    """
    try:
        rooms_url = typerconf.get(BOOKED_ROOMS_URL)
    except Exception as err:
        logging.error(f"Can't get URL from config: {err}")
        logging.info("Please set it with " "'nytid utils rooms set-url <url>'")
        raise typer.Exit(1)
    try:
        csv_out = csv.writer(sys.stdout, delimiter=delimiter)

        for start, end, rooms in booked_rooms(rooms_url):
            csv_out.writerow([start, end, ", ".join(sorted(rooms))])
    except Exception as err:
        logging.error(f"Can't get booked rooms: {err}")
        raise typer.Exit(1)


if __name__ == "__main__":
    cli()



import arrow.arrow as arrow
import calendar
import dateutil
import ics.event
import ics.icalendar
import requests
import re

TIMEEDIT_EVENTS = [
    "Datorlaboration",
    "Computer lab",
    "Examination",
    "Exkursion",
    "Excursion",
    "Fältövning",
    "Field exercise",
    "Frågestund",
    "Question time",
    "Question Time",
    "Föreläsning",
    "Lecture",
    "FX-komplettering",
    "FX-completion",
    "Handledning",
    "Tutorial",
    "Information",
    "Kontrollskrivning",
    "Partial Exam/Quiz",
    "Laboration",
    "Lab",
    "Lektion",
    "Lesson",
    "Omtenta",
    "Re-exam",
    "Re-Exam",
    "Presentation",
    "Projektarbete",
    "Project work",
    "Project Work",
    "Redovisning",
    "Reporting",
    "Räknestuga",
    "Math work session",
    "Seminarium",
    "Seminar",
    "Samverkansinlärning",
    "Supplemental instruction",
    "Supplemental Instruction",
    "Studiebesök",
    "Study visit",
    "Tentamen",
    "Exam",
    "Upprop",
    "Roll-call",
    "Webinarium",
    "Webinar",
    "Verksamhetsförlagd utbildning",
    "Pre-service Placement",
    "Workshop",
    "Övning",
    "Exercise",
]


def format_header(event, week=True):
    """
    Formats the event header, a one-line representation (week, date, time,
    event title)

    If the `week` parameter is `True` (default), the week number and weekday
    will be included in the header.
    """
    header = ""

    if week:
        header += (
            f"Week {event.begin.isocalendar()[1]} "
            + calendar.day_name[event.begin.weekday()]
            + " "
        )

    header += event.begin.to(dateutil.tz.tzlocal()).format("DD/MM HH:mm")

    header += f" {event.name}"

    return header


def format_event(event, week=True):
    """
    Takes event (ics.event.Event) object,
    returns long string representation (header, location, description over
    several lines)

    The `week` parameter is passed to the `format_header` function, it results
    in the week number and day being included in the header. Default is `True`.
    """
    header = format_header(event, week)
    location = event.location
    description = "\n".join(
        filter(lambda x: "http" not in x, event.description.splitlines())
    )

    return f"{header}\n{location}\n{description}".strip()


def filter_event_description(
    event_desc, ignore=["http", "grupp", "group", "ID ", "Daniel Bosk"], separator="; "
):
    """
    Takes event description event_desc as string,
    returns filtered string with newlines replaced by semicolons.
    Rows of description containing any term in list ignore, will not appear in
    the returned string.
    The filtered rows of the description are joined again by the string in
    separator.
    """
    desc_parts = event_desc.splitlines()
    desc_parts_keep = []

    for part in desc_parts:
        found = False
        for term in ignore:
            if term in part:
                found = True
                break

        if not found:
            desc_parts_keep.append(part)

    return separator.join(desc_parts_keep)


def format_event_short(event, week=False):
    """
    Takes event (ics.event.Event) object,
    returns a short string representation (one line)

    The `week` parameter is passed to the `format_header` function, it results
    in the week number and day being included in the header. Default is
    `False`.
    """
    header = format_header(event, week)
    description = filter_event_description(event.description)

    return f"{header} {description}".strip()


def format_event_csv(event, week=False, location=False):
    """
    Takes event (ics.event.Event) object,
    returns a list of strings, which can be used as a row in a CSV file.

      [0] Week number and day (if week is True)
      [1] Date and time
      [2] Event name
      [3] Location (if location is True)
      [4] Description

    The `week` parameter set to True results in the week number and day being
    included at index 0. Default is `False`.

    The `location` parameter set to True results in the location being included
    at index 2. Default is `False`.
    """
    row = []

    if week:
        row.append(
            f"Week {event.begin.isocalendar()[1]} "
            + calendar.day_name[event.begin.weekday()]
        )

    row.append(event.begin.to(dateutil.tz.tzlocal()).format("DD/MM HH:mm"))
    row.append(event.name)

    if location:
        row.append(event.location)

    description = filter_event_description(event.description, separator="; ")
    row.append(description)

    return row


def read_calendar(url):
    """
    Input: url is a string containing the URL to the ICS-formatted calendar.
    Output: an [[ics.icalendar.Calendar]] object.
    """
    response = requests.get(url)
    if response.status_code == requests.codes.ok:
        return ics.icalendar.Calendar(imports=response.text)
    raise Exception(response.text)


def event_filter(events, whitelisted=TIMEEDIT_EVENTS):
    """
    Takes a list of events (ics.event.Event), returns a filtered list of events
    (generator). The events to include are the teaching events.

    It covers the following events:

      "Datorlaboration", "Computer lab",
      "Examination",
      "Exkursion", "Excursion",
      "Fältövning", "Field exercise",
      "Frågestund", "Question time", "Question Time",
      "Föreläsning", "Lecture",
      "FX-komplettering", "FX-completion",
      "Handledning", "Tutorial",
      "Information",
      "Kontrollskrivning", "Partial Exam/Quiz",
      "Laboration", "Lab",
      "Lektion", "Lesson",
      "Omtenta", "Re-exam", "Re-Exam",
      "Presentation",
      "Projektarbete", "Project work", "Project Work",
      "Redovisning", "Reporting",
      "Räknestuga", "Math work session",
      "Seminarium", "Seminar",
      "Samverkansinlärning", "Supplemental instruction",
      "Supplemental Instruction",
      "Studiebesök", "Study visit",
      "Tentamen", "Exam",
      "Upprop", "Roll-call",
      "Webinarium", "Webinar",
      "Verksamhetsförlagd utbildning", "Pre-service Placement",
      "Workshop",
      "Övning", "Exercise"
    """
    for event in events:
        for event_type in whitelisted:
            if re.match(f"^{event_type}(,|\s|$)", event.name):
                yield event
                break



import os
import pathlib


class StorageRoot:
    """
    Abstracts the storage system. This allows us to use paths relative to a root.
    This of this as a simple chroot jail (just not for security).
    """

    def __init__(self, path: pathlib.Path):
        """
        Uses `path` as the root. All files opened are in subdirectories of `path`.
        """
        self.__path = pathlib.Path(path)
        self.__dir_fd = None

    @property
    def path(self) -> pathlib.Path:
        """
        `pathlib.Path` representation of the storage root.
        """
        return pathlib.Path(self.__path)

    def open(self, *args, **kwargs):
        """
        Takes the same arguments as built-in `open`.
        """
        return open(*args, **kwargs, opener=self.__opener)

    def __opener(self, path, flags):
        """The opener, used internally"""
        if not self.__dir_fd:
            self.open_dir()
        return os.open(path, flags, dir_fd=self.__dir_fd)

    def open_dir(self):
        """
        Open the storage root directory. Note that this is done automatically when
        opening a file using the `open` method.
        """
        try:
            self.__dir_fd = os.open(self.path, os.O_RDONLY)
        except FileNotFoundError:
            os.makedirs(self.path)
            self.__dir_fd = os.open(self.path, os.O_RDONLY)

    def close(self):
        """Close the storage root"""
        os.close(self.__dir_fd)
        self.__dir_fd = None

    def __enter__(self):
        """Enter of with statement"""
        self.open_dir()
        return self

    def __exit__(self, *args):
        """Exit of with statement"""
        self.close()

    def grant_access(user):
        """
        This method is not implemented for `nytid.cli.storage.StorageRoot`.
        """
        raise NotImplementedError

    def revoke_access(user):
        """
        This method is not implemented for `nytid.cli.storage.StorageRoot`.
        """
        raise NotImplementedError


def open_root(*args, **kwargs) -> StorageRoot:
    """
    Takes arguments (`*args` and `**kwargs`), determines which submodule is the
    best and passes the arguments onto the `StorageRoot` constructor of that
    module.
    """
    if args[0].startswith("/afs"):
        import nytid.storage.afs as storage_module
    if args[0].endswith(".git"):
        import nytid.storage.git as storage_module
    try:
        root = storage_module.StorageRoot(*args, **kwargs)
    except NameError:
        root = StorageRoot(*args, **kwargs)

    return root



"""The CLI of nytid"""

import typer
import importlib
import pathlib
import os
import typerconf

MODULE_EXTENSIONS = importlib.machinery.SOURCE_SUFFIXES

import logging
import sys

logging.basicConfig(format=f"nytid: %(levelname)s: %(message)s")

cli = typer.Typer(
    help="""
                       A CLI for managing TAs and courses.
                       """,
    epilog="Copyright (c) 2022--2025 Daniel Bosk, " "2022 Alexander Baltatzis.",
)


def package_contents(package_name, recurse=False):
    """
    Find all modules in a package. Recurse through subpackages if recurse is True
    (defualt False).
    """
    spec = importlib.util.find_spec(package_name)
    if spec is None:
        return set()

    pathname = pathlib.Path(spec.origin).parent

    modules = set()
    with os.scandir(pathname) as entries:
        for entry in entries:
            if entry.name.startswith("__"):
                continue
            current = ".".join((package_name, entry.name.partition(".")[0]))
            if entry.is_file():
                if any(
                    [entry.name.endswith(extension) for extension in MODULE_EXTENSIONS]
                ):
                    modules.add(current)
            elif entry.is_dir():
                modules.add(current)
                if recurse:
                    modules |= package_contents(current, recurse=True)

    return modules


modules = package_contents(__name__)
for module_name in modules:
    try:
        module = importlib.import_module(module_name)
        cli.add_typer(module.cli)
    except Exception as err:
        logging.warning(f"Trying to add {module_name} yields: {err}")
        continue
typerconf.add_config_cmd(cli)

if __name__ == "__main__":
    cli()
